@@ -30,29 +30,55 @@ void
 perl_simple_tag(context_p ctx, char **output, int argc, char **argv)
 {
     HV *perl_simple_tags = perl_get_hv(PERL_TAGS_SIMPLE_TAG_HASH, TRUE);
-    int i, n;
-    SV **coderef         = hv_fetch(perl_simple_tags, argv[0],
-                                    strlen(argv[0]), FALSE);
+    SV **coderef         = NULL;
     SV *perlcontext      = sv_newmortal();
+    context_p current    = ctx;
+    int i, retval;
+    char key[20];
     dSP;
     CV *code;
 
+
+    /* Build the unique key for this context */
+    while (current->parent_context != NULL)
+    {
+        current = current->parent_context;
+    }
+    snprintf(key, 20, "%p", current);
+
+
+    /* Create and bless a perl version of the context */
+    sv_setref_pv(perlcontext, "context_p", (void *)ctx);
+    sv_bless(perlcontext, gv_stashpv(TEMPLATE_PACKAGE, 0));
+
+
+    /* Fetch a code reference out of the perl_simple_tags hash */
+    if (hv_exists(perl_simple_tags, key, strlen(key)))
+    {
+        HV *stags = (HV *)SvRV(*(hv_fetch(perl_simple_tags,
+                                          key, strlen(key), FALSE)));
+        coderef = hv_fetch(stags, argv[0], strlen(argv[0]), FALSE);
+    }
+    else
+    {
+        *output = NULL;
+	return;
+    }
     if (coderef == NULL)
     {
         *output = NULL;
         return;
     }
-
     code = (CV *)*coderef;
 
+
     ENTER;
     SAVETMPS;
 
+
     /* push arguments */
     PUSHMARK(sp);
     /* first the context */
-    sv_setref_pv(perlcontext, "context_p", (void *)ctx);
-    sv_bless(perlcontext, gv_stashpv(TEMPLATE_PACKAGE, 0));
     XPUSHs(perlcontext);
     for (i = 0; i <= argc; i++)
     {
@@ -61,18 +87,22 @@ perl_simple_tag(context_p ctx, char **output, int argc, char **argv)
 
     PUTBACK;
 
-    n = perl_call_sv((SV *)code, G_SCALAR);
+    retval = perl_call_sv((SV *)code, G_SCALAR);
 
     SPAGAIN;
 
-    if (n == 1)
+
+    /* Translate the return into a char* for parser */
+    if (retval == 1)
     {
         char *t = POPp;
 
         *output = (char *)calloc(1, strlen(t) + 1);
         strncpy(*output, t, strlen(t));
         (*output)[strlen(t) + 1] = '\0';
-    } else {
+    }
+    else
+    {
         *output = NULL;
     }
 
@@ -98,28 +128,52 @@ void
 perl_tag_pair(context_p ctx, int argc, char **argv)
 {
     HV *perl_tag_pairs = perl_get_hv(PERL_TAGS_TAG_PAIR_HASH, TRUE);
-    int i, n;
-    SV **coderef         = hv_fetch(perl_tag_pairs, argv[0],
-                                    strlen(argv[0]), FALSE);
-    SV *perlcontext      = sv_newmortal();
+    SV **coderef       = NULL;
+    SV *perlcontext    = sv_newmortal();
+    context_p current  = ctx;
+    int i;
+    char key[20];
     dSP;
     CV *code;
 
+    /* Build the unique key for this context */
+    while (current->parent_context != NULL)
+    {
+        current = current->parent_context;
+    }
+    snprintf(key, 20, "%p", current);
+
+
+    /* Create and bless a perl version of the context */
+    sv_setref_pv(perlcontext, "context_p", (void *)ctx);
+    sv_bless(perlcontext, gv_stashpv(TEMPLATE_PACKAGE, 0));
+
+
+    /* Fetch a code reference out of the perl_tag_pairs hash */
+    if (hv_exists(perl_tag_pairs, key, strlen(key)))
+    {
+        HV *tagps = (HV *)SvRV(*(hv_fetch(perl_tag_pairs, key, strlen(key),
+                                          FALSE)));
+        coderef = hv_fetch(tagps, argv[0], strlen(argv[0]), FALSE);
+    }
+    else
+    {
+        return;
+    }
     if (coderef == NULL)
     {
         return;
     }
-
     code = (CV *)*coderef;
 
+
     ENTER;
     SAVETMPS;
 
+
     /* push arguments */
     PUSHMARK(sp);
     /* first the context */
-    sv_setref_pv(perlcontext, "context_p", (void *)ctx);
-    sv_bless(perlcontext, gv_stashpv(TEMPLATE_PACKAGE, 0));
     XPUSHs(perlcontext);
     for (i = 0; i <= argc; i++)
     {
@@ -128,7 +182,7 @@ perl_tag_pair(context_p ctx, int argc, char **argv)
 
     PUTBACK;
 
-    n = perl_call_sv((SV *)code, G_DISCARD);
+    perl_call_sv((SV *)code, G_DISCARD);
 
     SPAGAIN;
     PUTBACK;
